#'Compute Brier score and its decomposition and Brier skill score
#'
#'Compute the Brier score (BS) and the components of its standard decompostion
#'as well with the two within-bin components described in Stephenson et al., 
#'(2008). It also returns the bias-corrected decomposition of the BS (Ferro and
#'Fricker, 2012). BSS has the climatology as the reference forecast. 
#'
#'@param exp A vector or a numeric array with named dimensions of the probablistic  
#'  prediction data. The dimension must at least have 'time_dim'. It may have 
#'  'memb_dim' for performing ensemble mean. The values should be within the 
#'  range [0, 1].
#'@param obs A numeric array with named dimensions of the binary observations 
#'  (0 or 1). The dimension must at least have 'time_dim' and other dimensions 
#'  of 'exp' except 'memb_dim'.
#'@param thresholds A numeric vector used to bin the forecasts. The default 
#' value is \code{seq(0, 1, 0,1)}, which means that the bins are 
#'  \code{[0,0.1), [0.1, 0.2), ... [0.9, 1]}.
#'@param time_dim A character string indicating the name of dimension along  
#'  which Brier score is computed. The default value is 'sdate'.
#'@param memb_dim A character string of the name of the member dimension. It 
#'  must be one dimension of 'exp'. The function will do the ensemble mean 
#'  over this dimension. If there is no member dimension, set NULL. The default
#'  value is NULL.
#'
#'@return A list that contains:
#'The numeric arrays with all 'exp' and 'obs' dimensions expect 'time_dim' and
#''memb_dim': 
#'\item{$rel}{standard reliability}
#'\item{$res}{standard resolution}
#'\item{$unc}{standard uncertainty}  
#'\item{$bs}{Brier score}
#'\item{$bs_check_res}{rel - res + unc}
#'\item{$bss_res}{res - rel / unc}
#'\item{$gres}{generalized resolution}
#'\item{$bs_check_gres}{rel - gres + unc}
#'\item{$bss_gres}{gres - rel / unc}
#'\item{$rel_bias_corrected}{bias - corrected rel}
#'\item{$gres_bias_corrected}{bias - corrected gres}
#'\item{$unc_bias_corrected}{bias - corrected unc}
#'\item{$bss_bias_corrected}{gres_bias_corrected - rel_bias_corrected / unc_bias_corrected}
#'The numeric arrays with the same dimensions as above and one additional
#'dimension 'bin':
#'\item{$nk}{number of forecast in each bin}
#'\item{$fkbar}{average probability of each bin}
#'\item{$okbar}{relative frequency that the observed event occurred}
#'
#'@references
#'Wilks (2006) Statistical Methods in the Atmospheric Sciences.\cr
#'Stephenson et al. (2008). Two extra components in the Brier score decomposition. 
#'  Weather and Forecasting, 23: 752-757.\cr
#'Ferro and Fricker (2012). A bias-corrected decomposition of the BS. 
#'  Quarterly Journal of the Royal Meteorological Society, DOI: 10.1002/qj.1924.
#'
#'@examples
#'# Inputs are vectors
#'exp <- runif(10)
#'obs <- round(a)
#'x <- BrierScore(exp, obs)
#'res <- x$bs - x$bs_check_res
#'res <- x$bs - x$bs_check_gres
#'res <- x$rel_bias_corrected - x$gres_bias_corrected + x$unc_bias_corrected
#'
#'# Inputs are arrays
#'example(Load)
#'bins_ano_exp <- ProbBins(ano_exp, thr = c(1/3, 2/3), posdates = 3, posdim = 2)
#'bins_ano_obs <- ProbBins(ano_obs, thr = c(1/3, 2/3), posdates = 3, posdim = 2)
#'res <- BrierScore(bins_ano_exp, MeanDims(bins_ano_obs, 'member'), memb_dim = 'member') 
#'
#'@import multiApply
#'@export
BrierScore <- function(exp, obs, thresholds = seq(0, 1, 0.1), time_dim = 'sdate',
                       memb_dim = NULL, ncores = NULL) {

  # Check inputs
  ## exp and obs
  if (is.null(exp) | is.null(obs)) {
    stop("Parameter 'exp' and 'obs' cannot be NULL.")
  }
  if (!is.numeric(exp) | !is.numeric(obs)) {
    stop("Parameter 'exp' and 'obs' must be a vector or a numeric array.")
  }
  if (is.null(dim(exp))) {  #is vector
    dim(exp) <- c(length(exp))
    names(dim(exp)) <- time_dim
  }
  if (is.null(dim(obs))) {  #is vector
    dim(obs) <- c(length(obs))
    names(dim(obs)) <- time_dim
  }
  if(any(is.null(names(dim(exp))))| any(nchar(names(dim(exp))) == 0) |
     any(is.null(names(dim(obs))))| any(nchar(names(dim(obs))) == 0)) {
    stop("Parameter 'exp' and 'obs' must have dimension names.")
  }
  if (max(exp) > 1 | min(exp) < 0) {
    stop("Parameter 'exp' must be within [0, 1] range.")
  }
  if (any(!obs %in% c(0, 1))) {
    stop("Parameter 'obs' must be binary events (0 or 1).")
  }
  name_exp <- sort(names(dim(exp)))
  name_obs <- sort(names(dim(obs)))
  if (!is.null(memb_dim)) {
    name_exp <- name_exp[-which(name_exp == memb_dim)]
  }
  if (any(name_exp != name_obs)) {
    stop(paste0("Parameter 'exp' and 'obs' must have the same names and lengths ",
                "of all the dimensions expect 'exp' may have 'memb_dim'."))
  }
  if (!all(dim(exp)[name_exp] == dim(obs)[name_obs])) {
    stop(paste0("Parameter 'exp' and 'obs' must have the same names and lengths ",
                "of all the dimensions expect 'exp' may have 'memb_dim'."))
  }
  ## thresholds
  if (!is.numeric(thresholds) | !is.vector(thresholds)) {
    stop("Parameter 'thresholds' must be a numeric vector.")
  }
  ## time_dim
  if (!is.character(time_dim) | length(time_dim) > 1) {
    stop("Parameter 'time_dim' must be a character string.")
  }
  if (!time_dim %in% names(dim(exp))) {
    stop("Parameter 'time_dim' is not found in 'exp' and 'obs' dimension.")
  }
  ## memb_dim
  if (!is.null(memb_dim)) {
    if (!is.character(memb_dim) | length(memb_dim) > 1) {
      stop("Parameter 'memb_dim' must be a character string.")
    }
    if (!memb_dim %in% names(dim(exp))) {
      stop("Parameter 'memb_dim' is not found in 'exp' dimension.")
    }
  }
  ## ncores
  if (!is.null(ncores)) {
    if (!is.numeric(ncores) | ncores %% 1 != 0 | ncores <= 0 |
      length(ncores) > 1) {
      stop("Parameter 'ncores' must be a positive integer.")
    }
  } 

  ###############################
  # Calculate Brier score

  ## ensemble mean
  if (!is.null(memb_dim)) {
    exp <- MeanDims(exp, memb_dim)
  }

  res <- Apply(list(exp, obs), 
               target_dims = list(c(time_dim), 
                                  c(time_dim)),
               fun = .BrierScore, 
               thresholds = thresholds,
               ncores = ncores)

 return(res)
}

.BrierScore <- function(exp, obs, thresholds = seq(0, 1, 0.1)) {

  # exp: [sdate]
  # obs: [sdate]

  n <- length(exp)
  nbins <- length(thresholds) - 1  # Number of bins
  bins <- as.list(paste("bin", 1:nbins, sep = ""))
  for (i in 1:nbins) {
    if (i == nbins) {
      bins[[i]] <- list(which(exp >= thresholds[i] & exp <= thresholds[i + 1]))
    } else {
      bins[[i]] <- list(which(exp >= thresholds[i] & exp < thresholds[i + 1]))
    }
  }

  fkbar <- okbar <- nk <- array(0, dim = nbins)
  for (i in 1:nbins) {
    nk[i] <- length(bins[[i]][[1]])
    fkbar[i] <- sum(exp[bins[[i]][[1]]]) / nk[i]
    okbar[i] <- sum(obs[bins[[i]][[1]]]) / nk[i]
  }
    
#-----in old .BrierScore()---------
#    fkbar[fkbar == Inf] <- 0 
#    okbar[is.nan(okbar)] <- 0
#----------------------------------

  obar <- sum(obs) / length(obs)
  relsum <- ressum <- term1 <- term2 <- 0
  for (i in 1:nbins) {
    if (nk[i] > 0) {
      relsum <- relsum + nk[i] * (fkbar[i] - okbar[i])^2
      ressum <- ressum + nk[i] * (okbar[i] - obar)^2
      for (j in 1:nk[i]) {
        term1 <- term1 + (exp[bins[[i]][[1]][j]] - fkbar[i])^2
        term2 <- term2 + (exp[bins[[i]][[1]][j]] - fkbar[i]) * (obs[bins[[i]][[1]][j]] - okbar[i])
      }
    }
  }
  rel <- relsum / n
  res <- ressum / n
  unc <- obar * (1 - obar)
  bs <- sum((exp - obs)^2) / n
  bs_check_res <- rel - res + unc
  bss_res <- (res - rel) / unc
  gres <- res - term1 * (1 / n) + term2 * (2 / n)   # Generalized resolution
  bs_check_gres <- rel - gres + unc                 # BS using GRES
  bss_gres <- (gres - rel) / unc                    # BSS using GRES
    
  
  # Estimating the bias-corrected components of the BS 
  term3 <- array(0, nbins)
  for (i in 1:nbins) {
    term3[i] <- (nk[i] / (nk[i] - 1)) * okbar[i] * (1 - okbar[i])
  }
  term_a <- sum(term3,  na.rm = T) / n
  term_b <- (obar * (1 - obar)) / (n - 1)
  rel_bias_corrected <- rel - term_a
  gres_bias_corrected <- gres - term_a + term_b
  if (rel_bias_corrected < 0 || gres_bias_corrected < 0) {
    rel_bias_corrected2 <- max(rel_bias_corrected, rel_bias_corrected - gres_bias_corrected, 0)
    gres_bias_corrected2 <- max(gres_bias_corrected, gres_bias_corrected - rel_bias_corrected, 0)
    rel_bias_corrected <- rel_bias_corrected2
    gres_bias_corrected <- gres_bias_corrected2
  }
  unc_bias_corrected <- unc + term_b
  bss_bias_corrected <- (gres_bias_corrected - rel_bias_corrected) / unc_bias_corrected
  
  #if (round(bs, 8) == round(bs_check_gres, 8) & round(bs_check_gres, 8) == round((rel_bias_corrected - gres_bias_corrected + unc_bias_corrected), 8)) {
  #  cat("No error found  \ n")
  #  cat("BS = REL - GRES + UNC = REL_lessbias - GRES_lessbias + UNC_lessbias  \ n")
  #}
   
  # Add name for nk, fkbar, okbar
  names(dim(nk)) <- 'bin'
  names(dim(fkbar)) <- 'bin'
  names(dim(okbar)) <- 'bin'

  res_list <- list(rel = rel, res = res, unc = unc, bs = bs, bs_check_res = bs_check_res,
                   bss_res = bss_res, gres = gres, bs_check_gres = bs_check_gres,
                   bss_gres = bss_gres, rel_bias_corrected = rel_bias_corrected,
                   gres_bias_corrected = gres_bias_corrected,
                   unc_bias_corrected = unc_bias_corrected,
                   bss_bias_corrected = bss_bias_corrected, nk = nk, fkbar = fkbar,
                   okbar = okbar) #bins = list(bins), 

  return(invisible(res_list))
}
